// Copyright 2009 Kwiirk
// Licensed under the terms of the GNU GPL, version 2
// http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt


#include "libwbfs.h"

#ifdef __GNUC__
#define likely(x)       __builtin_expect(!!(x), 1)
#define unlikely(x)     __builtin_expect(!!(x), 0)
#else if _MSC_VER
#define likely(x)       (!!(x))
#define unlikely(x)     (!!(x))
#endif

#define ERROR_WBFS(x) do {wbfs_error(x);goto error;}while(0)
#define ALIGN_LBA(x) (((x)+p->hd_sec_sz-1)&(~(p->hd_sec_sz-1)))
static int force_mode=0;
void wbfs_set_force_mode(int force)
{
	force_mode = force;
}
static unsigned char size_to_shift(unsigned int size)
{
	unsigned char ret = 0;
	while(size)
	{
		ret++;
		size>>=1;
	}
	return ret-1;
}
#define read_le32_unaligned(x) ((x)[0]|((x)[1]<<8)|((x)[2]<<16)|((x)[3]<<24))

wbfs_t*wbfs_open_hd(rw_sector_callback_t read_hdsector,
					rw_sector_callback_t write_hdsector,
					void *callback_data,
					int hd_sector_size, unsigned int num_hd_sector, int reset)
{
	unsigned int i=num_hd_sector,ret;
	unsigned char *ptr,*tmp_buffer = (unsigned char*)wbfs_ioalloc(hd_sector_size);
	unsigned char part_table[16*4];
	ret = read_hdsector(callback_data,0,1,tmp_buffer);
	if(ret)
		return 0;
	//find wbfs partition
	wbfs_memcpy(part_table,tmp_buffer+0x1be,16*4);
	ptr = part_table;
	for(i=0;i<4;i++,ptr+=16)
	{
		unsigned int part_lba = read_le32_unaligned(ptr+0x8);
		wbfs_head_t *head = (wbfs_head_t *)tmp_buffer;
		ret = read_hdsector(callback_data,part_lba,1,tmp_buffer);
		// verify there is the magic.
		if (head->magic == wbfs_htonl(WBFS_MAGIC))
		{
			wbfs_t*p = wbfs_open_partition(read_hdsector,write_hdsector,
				callback_data,0,hd_sector_size,0,part_lba,reset);
			return p;
		}
	}
	if(reset)// XXX make a empty hd partition..
	{
	}
	return 0;
}
wbfs_t*wbfs_open_partition(rw_sector_callback_t read_hdsector,
						   rw_sector_callback_t write_hdsector,
						   void *callback_data, int offset,
						   int hd_sector_size, unsigned int num_hd_sector, unsigned int part_lba, int reset)
{
	scopeHANDLE handle = callback_data;
	wbfs_t *p = (wbfs_t*)wbfs_malloc(sizeof(wbfs_t));

	wbfs_head_t *head = (wbfs_head_t*)wbfs_ioalloc(hd_sector_size?hd_sector_size:512);

	//constants, but put here for consistancy
	p->wii_sec_sz = 0x8000;
	p->wii_sec_sz_s = size_to_shift(0x8000);
	p->n_wii_sec = (num_hd_sector/0x8000)*hd_sector_size;
	p->n_wii_sec_per_disc = 143432*2;//support for double layers discs..
	p->head = head;
	p->part_lba = part_lba;
	p->reading_offset = offset;
	// init the partition
	if (reset)
	{
		unsigned char sz_s;
		wbfs_memset(head,0,hd_sector_size);
		head->magic = wbfs_htonl(WBFS_MAGIC);
		head->hd_sec_sz_s = size_to_shift(hd_sector_size);
		head->n_hd_sec = wbfs_htonl(num_hd_sector);
		// choose minimum wblk_sz that fits this partition size
		for(sz_s=6;sz_s<11;sz_s++)
		{
			// ensure that wbfs_sec_sz is big enough to address every blocks using 16 bits
			if(p->n_wii_sec <((1U<<16)*(1<<sz_s)))
				break;
		}
		head->wbfs_sec_sz_s = sz_s+p->wii_sec_sz_s;
	}else
		read_hdsector(callback_data,p->part_lba + p->reading_offset,1,head);
	if (head->magic != wbfs_htonl(WBFS_MAGIC))
		ERROR_WBFS("bad magic");
	if(!force_mode && hd_sector_size && head->hd_sec_sz_s !=  size_to_shift(hd_sector_size))
		ERROR_WBFS("hd sector size doesn't match");
	if(!force_mode && num_hd_sector && head->n_hd_sec != wbfs_htonl(num_hd_sector))
		ERROR_WBFS("hd num sector doesn't match");
	p->hd_sec_sz = 1<<head->hd_sec_sz_s;
	p->hd_sec_sz_s = head->hd_sec_sz_s;
	p->n_hd_sec = wbfs_ntohl(head->n_hd_sec);

	p->n_wii_sec = (p->n_hd_sec/p->wii_sec_sz)*(p->hd_sec_sz);

	p->wbfs_sec_sz_s = head->wbfs_sec_sz_s;
	p->wbfs_sec_sz = 1<<p->wbfs_sec_sz_s;
	p->n_wbfs_sec = p->n_wii_sec >> (p->wbfs_sec_sz_s - p->wii_sec_sz_s);
	p->n_wbfs_sec_per_disc = p->n_wii_sec_per_disc >> (p->wbfs_sec_sz_s - p->wii_sec_sz_s);
	p->disc_info_sz = ALIGN_LBA(sizeof(wbfs_disc_info_t) + p->n_wbfs_sec_per_disc*2);

	//printf("hd_sector_size %X wii_sector size %X wbfs sector_size %X\n",p->hd_sec_sz,p->wii_sec_sz,p->wbfs_sec_sz);
	p->read_hdsector = read_hdsector;
	p->write_hdsector = write_hdsector;
	p->callback_data = callback_data;
	handle = INVALID_HANDLE_VALUE;

	p->freeblks_lba = (p->wbfs_sec_sz - p->n_wbfs_sec/8)>>p->hd_sec_sz_s;

	if(!reset)
		p->freeblks = 0; // will alloc and read only if needed
	else
	{
		// init with all free blocks
		p->freeblks = (unsigned int*)wbfs_ioalloc(ALIGN_LBA(p->n_wbfs_sec/8));
		wbfs_memset(p->freeblks,0xff,p->n_wbfs_sec/8);
	}
	p->max_disc = (p->freeblks_lba-1)/(p->disc_info_sz>>p->hd_sec_sz_s);
	if(p->max_disc > p->hd_sec_sz - sizeof(wbfs_head_t))
		p->max_disc = p->hd_sec_sz - sizeof(wbfs_head_t);

	p->tmp_buffer = (unsigned char*)wbfs_ioalloc(p->hd_sec_sz);
	p->n_disc_open = 0;

	return p;
error:
	wbfs_free(p);
	wbfs_iofree(head);
	return 0;

}

void wbfs_sync(wbfs_t*p)
{
	// copy back descriptors
	if(p->write_hdsector){
		p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+0,1, p->head);

		if(p->freeblks)
			p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+p->freeblks_lba,ALIGN_LBA(p->n_wbfs_sec/8)>>p->hd_sec_sz_s, p->freeblks);
	}
}
void wbfs_close(wbfs_t*p)
{
	wbfs_sync(p);

	if(p->n_disc_open)
		ERROR_WBFS("trying to close wbfs while discs still open");

	wbfs_iofree(p->head);
	wbfs_iofree(p->tmp_buffer);
	if(p->freeblks)
		wbfs_iofree(p->freeblks);

	wbfs_free(p);

error:
	return;
}

wbfs_disc_t *wbfs_open_disc(wbfs_t* p, unsigned char *discid)
{
	unsigned int i;
	unsigned int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;
	wbfs_disc_t *d = 0;
	for(i=0;i<p->max_disc;i++)
	{
		if (p->head->disc_table[i]){
			p->read_hdsector(p->callback_data,
				p->reading_offset + p->part_lba+1+i*disc_info_sz_lba,1,p->tmp_buffer);
			if(wbfs_memcmp(discid,p->tmp_buffer,6)==0){
				d = (wbfs_disc_t*)wbfs_malloc(sizeof(*d));
				if(!d)
					ERROR_WBFS("allocating memory");
				d->p = p;
				d->i = i;
				d->header = (wbfs_disc_info_t*)wbfs_ioalloc(p->disc_info_sz);
				if(!d->header)
					ERROR_WBFS("allocating memory");
				p->read_hdsector(p->callback_data,
					p->reading_offset + p->part_lba+1+i*disc_info_sz_lba,
					disc_info_sz_lba,d->header);
				p->n_disc_open ++;

				return d;
			}
		}
	}
	return 0;
error:
	if(d)
		wbfs_iofree(d);
	return 0;

}
void wbfs_close_disc(wbfs_disc_t*d)
{
	d->p->n_disc_open --;
	wbfs_iofree(d->header);
	wbfs_free(d);
}
// offset is pointing 32bit words to address the whole dvd, although len is in bytes
int wbfs_disc_read(wbfs_disc_t*d,unsigned int offset, unsigned char *data, unsigned int len)
{
	wbfs_t *p = d->p;
	unsigned short wlba = offset>>(p->wbfs_sec_sz_s-2);
	unsigned int iwlba_shift = p->wbfs_sec_sz_s - p->hd_sec_sz_s;
	unsigned int lba_mask = (p->wbfs_sec_sz-1)>>(p->hd_sec_sz_s);
	unsigned int lba = (offset>>(p->hd_sec_sz_s-2))&lba_mask;
	unsigned int off = offset&((p->hd_sec_sz>>2)-1);
	unsigned short iwlba = wbfs_ntohs(d->header->wlba_table[wlba]);
	unsigned int len_copied;
	int err = 0;
	unsigned char  *ptr = data;
	if(unlikely(iwlba==0))
		return 1;
	if(unlikely(off)){
		off*=4;
		err = p->read_hdsector(p->callback_data,
			p->reading_offset + p->part_lba + (iwlba<<iwlba_shift) + lba, 1, p->tmp_buffer);
		if(err)
			return err;
		len_copied = p->hd_sec_sz - off;
		if(likely(len < len_copied))
			len_copied = len;
		wbfs_memcpy(ptr, p->tmp_buffer + off, len_copied);
		len -= len_copied;
		ptr += len_copied;
		lba++;
		if(unlikely(lba>lba_mask && len)){
			lba=0;
			iwlba =  wbfs_ntohs(d->header->wlba_table[++wlba]);
			if(unlikely(iwlba==0))
				return 1;
		}
	}
	while(likely(len>=p->hd_sec_sz))
	{
		unsigned int nlb = len>>(p->hd_sec_sz_s);

		if(unlikely(lba + nlb > p->wbfs_sec_sz)) // dont cross wbfs sectors..
			nlb = p->wbfs_sec_sz-lba;
		err = p->read_hdsector(p->callback_data,
			p->reading_offset + p->part_lba + (iwlba<<iwlba_shift) + lba, nlb, ptr);
		if(err)
			return err;
		len -= nlb<<p->hd_sec_sz_s;
		ptr += nlb<<p->hd_sec_sz_s;
		lba += nlb;
		if(unlikely(lba>lba_mask && len)){
			lba = 0;
			iwlba =wbfs_ntohs(d->header->wlba_table[++wlba]);
			if(unlikely(iwlba==0))
				return 1;
		}
	}
	if(unlikely(len)){
		err = p->read_hdsector(p->callback_data,
			p->reading_offset + p->part_lba + (iwlba<<iwlba_shift) + lba, 1, p->tmp_buffer);
		if(err)
			return err;
		wbfs_memcpy(ptr, p->tmp_buffer, len);
	}     
	return 0;
}

// disc listing
unsigned int wbfs_count_discs(wbfs_t*p)
{
	unsigned int i,count=0;
	for(i=0;i<p->max_disc;i++)
		if (p->head->disc_table[i])
			count++;
	return count;

}
unsigned int wbfs_sector_used(wbfs_t *p,wbfs_disc_info_t *di)
{
	unsigned int tot_blk=0,j;
	for(j=0;j<p->n_wbfs_sec_per_disc;j++)
		if(wbfs_ntohs(di->wlba_table[j]))
			tot_blk++;
	return tot_blk;

}
unsigned int wbfs_get_disc_info(wbfs_t*p, unsigned int index,unsigned char *header,int header_size,unsigned int *size)//size in 32 bit
{
	unsigned int i,count=0;
	unsigned int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;
	for(i=0;i<p->max_disc;i++)
		if (p->head->disc_table[i]){
			if(count++==index)
			{
				p->read_hdsector(p->callback_data,
					p->reading_offset + p->part_lba+1+i*disc_info_sz_lba,1,p->tmp_buffer);
				if(header_size > (int)p->hd_sec_sz)
					header_size = p->hd_sec_sz;
				unsigned int magic = wbfs_ntohl(*(unsigned int*)(p->tmp_buffer+24));
				if(magic!=0x5D1C9EA3){
					p->head->disc_table[i]=0;
					return 1;
				}
				memcpy(header,p->tmp_buffer,header_size);
				if(size)
				{
					unsigned char *header = (unsigned char*)wbfs_ioalloc(p->disc_info_sz);
					p->read_hdsector(p->callback_data,
						p->reading_offset + p->part_lba+1+i*disc_info_sz_lba,disc_info_sz_lba,header);
					unsigned int sec_used = wbfs_sector_used(p,(wbfs_disc_info_t *)header);
					wbfs_iofree(header);
					*size = sec_used<<(p->wbfs_sec_sz_s-2);
				}
				return 0;
			}
		}
		return 1;
}

static void load_freeblocks(wbfs_t*p)
{
	if(p->freeblks)
		return;
	// XXX should handle malloc error..
	p->freeblks = (unsigned int*)wbfs_ioalloc(ALIGN_LBA(p->n_wbfs_sec/8));
	p->read_hdsector(p->callback_data,p->reading_offset + p->part_lba+p->freeblks_lba,ALIGN_LBA(p->n_wbfs_sec/8)>>p->hd_sec_sz_s, p->freeblks);

}
unsigned int wbfs_count_usedblocks(wbfs_t*p)
{
	unsigned int j,count=0;
	load_freeblocks(p);
	for(int i=0;i<p->n_wbfs_sec/(8*4);i++)
	{
		unsigned int v = wbfs_ntohl(p->freeblks[i]);
		if(v == ~0U)
			count+=32;
		else if(v!=0)
			for(j=0;j<32;j++)
				if (v & (1<<j))
					count++;
	}
	return count;
}


// write access


static int block_used(unsigned char *used,unsigned int i,unsigned int wblk_sz)
{
	unsigned int k;
	i*=wblk_sz;
	for(k=0;k<wblk_sz;k++)
		if(i+k<143432*2 && used[i+k])
			return 1;
	return 0;
}

static unsigned int alloc_block(wbfs_t*p)
{
	for(int i=0;i<p->n_wbfs_sec/(8*4);i++)
	{
		unsigned int v = wbfs_ntohl(p->freeblks[i]);
		if(v != 0)
		{
			for(int j=0;j<32;j++)
				if (v & (1<<j))
				{
					p->freeblks[i] = wbfs_htonl(v & ~(1<<j));
					return (i*32)+j+1;
				}
		}
	}
	return ~0;
}
static void free_block(wbfs_t *p,int bl)
{
	int i = (bl-1)/(32);
	int j = (bl-1)&31;
	unsigned int v = wbfs_ntohl(p->freeblks[i]);
	p->freeblks[i] = wbfs_htonl(v | 1<<j);
}
unsigned int wbfs_add_disc(wbfs_t*p,read_wiidisc_callback_t read_src_wii_disc,
						   void *callback_data,progress_callback_t spinner,partition_selector_t sel,int copy_1_1)
{
	int i,discn;
	unsigned int tot,cur;
	unsigned int wii_sec_per_wbfs_sect = 1<<(p->wbfs_sec_sz_s-p->wii_sec_sz_s);
	wiidisc_t *d = 0;
	unsigned char *used = 0;
	wbfs_disc_info_t *info = 0;
	unsigned char* copy_buffer = 0;
	used = (unsigned char*)wbfs_malloc(p->n_wii_sec_per_disc);
	if(!used)
		ERROR_WBFS("unable to alloc memory");
	if(!copy_1_1)
	{
		d = wd_open_disc(read_src_wii_disc,callback_data);
		if(!d)
			ERROR_WBFS("unable to open wii disc");
		wd_build_disc_usage(d,sel,used);
		wd_close_disc(d);
		d = 0;
	}

	for(i=0;i<p->max_disc;i++)// find a free slot.
		if(p->head->disc_table[i]==0)
			break;
	if(i==p->max_disc)
		ERROR_WBFS("no space left on device (table full)");
	p->head->disc_table[i] = 1;
	discn = i;
	load_freeblocks(p);

	// build disc info
	info = (wbfs_disc_info_t*)wbfs_ioalloc(p->disc_info_sz);
	read_src_wii_disc(callback_data,0,0x100,info->disc_header_copy);

	copy_buffer = (unsigned char*)wbfs_ioalloc(p->wii_sec_sz);
	if(!copy_buffer)
		ERROR_WBFS("alloc memory");
	tot=0;
	cur=0;
	if(spinner){
		// count total number to write for spinner
		for(i=0; i<p->n_wbfs_sec_per_disc;i++)
		{
			if(copy_1_1 || block_used(used,i,wii_sec_per_wbfs_sect)) tot += wii_sec_per_wbfs_sect;
		}
		spinner(0,tot);
	}
	for(i=0; i<p->n_wbfs_sec_per_disc;i++){
		unsigned short bl = 0;
		if(copy_1_1 || block_used(used,i,wii_sec_per_wbfs_sect)) {
			unsigned short j;

			bl = alloc_block(p);
			if (bl==0xffff)
				ERROR_WBFS("no space left on device (disc full)");
			for(j=0; j<wii_sec_per_wbfs_sect;j++) {
				unsigned int offset = (i*(p->wbfs_sec_sz>>2)) + (j*(p->wii_sec_sz>>2));

				read_src_wii_disc(callback_data,offset,p->wii_sec_sz,copy_buffer);

				//fix the partition table
				if(offset == (0x40000>>2))
					wd_fix_partition_table(d, sel, copy_buffer);
				p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+bl*(p->wbfs_sec_sz/p->hd_sec_sz)+j*(p->wii_sec_sz/p->hd_sec_sz),
					p->wii_sec_sz/p->hd_sec_sz,copy_buffer);
				cur++;
				if(spinner)
					spinner(cur,tot);
			}
		}
		info->wlba_table[i] = wbfs_htons(bl);
	}
	// write disc info
	int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;
	p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+1+discn*disc_info_sz_lba,disc_info_sz_lba,info);
	wbfs_sync(p);
error:
	if(d)
		wd_close_disc(d);
	if(used)
		wbfs_free(used);
	if(info)
		wbfs_iofree(info);
	if(copy_buffer)
		wbfs_iofree(copy_buffer);
	// init with all free blocks

	return 0;
}

unsigned int wbfs_add_ciso_disc(wbfs_t*p,read_wiidisc_callback_t read_src_wii_disc,
								void *callback_data,progress_callback_t spinner,partition_selector_t sel,int copy_1_1)
{
	int i,discn;
	unsigned int wii_sec_per_wbfs_sect = 1<<(p->wbfs_sec_sz_s-p->wii_sec_sz_s);
	unsigned char *used = 0;
	wbfs_disc_info_t *info = 0;
	unsigned char* copy_buffer = 0;
	used = (unsigned char*)wbfs_malloc(p->n_wii_sec_per_disc);

	if(!used)
		ERROR_WBFS("unable to alloc memory");

	for(i=0;i<p->max_disc;i++)// find a free slot.
		if(p->head->disc_table[i]==0)
			break;

	if(i==p->max_disc)
		ERROR_WBFS("no space left on device (table full)");

	p->head->disc_table[i] = 1;
	discn = i;
	load_freeblocks(p);

	copy_buffer = (unsigned char*)wbfs_ioalloc(p->wii_sec_sz);
	if(!copy_buffer)
		ERROR_WBFS("alloc memory");

	int reading_offset = 0;
	read_src_wii_disc(callback_data, reading_offset, 0x8000, copy_buffer);
	reading_offset += 0x8000;

	int block_size = (copy_buffer[7]<<24) + (copy_buffer[6]<<16) + (copy_buffer[5]<<8) + copy_buffer[4];

	for( i=8; i< 0x8000; i++)
	{
		if (copy_buffer[i]);
	}

	// build disc info
	info = (wbfs_disc_info_t*)wbfs_ioalloc(p->disc_info_sz);
	read_src_wii_disc(callback_data,reading_offset,0x100,info->disc_header_copy);

	for(i=0; i<p->n_wbfs_sec_per_disc;i++){
		unsigned short bl = 0;
		if(copy_1_1 || block_used(used,i,wii_sec_per_wbfs_sect)) {
			unsigned short j;

			bl = alloc_block(p);
			if (bl==0xffff)
				ERROR_WBFS("no space left on device (disc full)");
			for(j=0; j<wii_sec_per_wbfs_sect;j++) {
				unsigned int offset = (i*(p->wbfs_sec_sz>>2)) + (j*(p->wii_sec_sz>>2));

				read_src_wii_disc(callback_data,offset,p->wii_sec_sz,copy_buffer);

				p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+bl*(p->wbfs_sec_sz/p->hd_sec_sz)+j*(p->wii_sec_sz/p->hd_sec_sz),
					p->wii_sec_sz/p->hd_sec_sz,copy_buffer);
			}
		}
				if(spinner)
					spinner(i,p->n_wbfs_sec_per_disc);

		info->wlba_table[i] = wbfs_htons(bl);
	}
	// write disc info
	int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;
	p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+1+discn*disc_info_sz_lba,disc_info_sz_lba,info);
	wbfs_sync(p);
error:
	if(used)
		wbfs_free(used);
	if(info)
		wbfs_iofree(info);
	if(copy_buffer)
		wbfs_iofree(copy_buffer);
	// init with all free blocks

	return 0;
}
unsigned int wbfs_rm_disc(wbfs_t*p, unsigned char* discid)
{
	wbfs_disc_t *d = wbfs_open_disc(p,discid);
	int i;
	int discn = 0;
	int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;
	if(!d)
		return 1;
	load_freeblocks(p);
	discn = d->i;
	for( i=0; i< p->n_wbfs_sec_per_disc; i++)
	{
		unsigned int iwlba = wbfs_ntohs(d->header->wlba_table[i]);
		if (iwlba)
			free_block(p,iwlba);
	}
	memset(d->header,0,p->disc_info_sz);
	p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+1+discn*disc_info_sz_lba,disc_info_sz_lba,d->header);
	p->head->disc_table[discn] = 0;
	wbfs_close_disc(d);
	wbfs_sync(p);
	return 0;
}

// trim the file-system to its minimum size
unsigned int wbfs_trim(wbfs_t*p);

unsigned int wbfs_copy_disc(wbfs_disc_t*d_src, wbfs_t*p_dst, progress_callback_t spinner)
{
	wbfs_t *p_src = d_src->p;
	unsigned char* copy_buffer = 0;
	unsigned char* dst_copy_buffer = 0;
	int i;

	unsigned int src_wbs_nlb=p_src->wbfs_sec_sz/p_src->hd_sec_sz;
	unsigned int dst_wbs_nlb=p_dst->wbfs_sec_sz/p_dst->hd_sec_sz;
	wbfs_disc_info_t *info = NULL;

	if( src_wbs_nlb%dst_wbs_nlb != 0 && dst_wbs_nlb%src_wbs_nlb != 0)
	{
		ERROR_WBFS("Difference between source and dest");
	}

	int n_src_in_dst = 1;
	int n_dst_in_src = 1;

	if( src_wbs_nlb > dst_wbs_nlb )
	{
		n_dst_in_src = src_wbs_nlb / dst_wbs_nlb;
	}
	else
	{
		n_src_in_dst = dst_wbs_nlb / src_wbs_nlb;
	}

	copy_buffer = (unsigned char*)wbfs_ioalloc(p_src->wbfs_sec_sz);
	if(!copy_buffer)
		ERROR_WBFS("alloc memory");

	for(i=0;i<p_dst->max_disc;i++)// find a free slot.
		if(p_dst->head->disc_table[i]==0)
			break;
	if(i==p_dst->max_disc)
		ERROR_WBFS("no space left on device (table full)");
	p_dst->head->disc_table[i] = 1;
	int discn = i;
	load_freeblocks(p_dst);

	info = (wbfs_disc_info_t*)wbfs_ioalloc(p_dst->disc_info_sz);
	memset(info, 0, p_dst->disc_info_sz);

	memcpy(info->disc_header_copy, d_src->header->disc_header_copy, 0x100);

	if( src_wbs_nlb == dst_wbs_nlb)
	{
		for( i=0; i< p_src->n_wbfs_sec_per_disc; i++)
		{
			unsigned int iwlba = wbfs_ntohs(d_src->header->wlba_table[i]);
			unsigned short bl = 0;
			if (iwlba)
			{
				bl = alloc_block(p_dst);
				if (bl==0xffff)
					ERROR_WBFS("no space left on device (disc full)");

				p_src->read_hdsector( p_src->callback_data, p_src->reading_offset + p_src->part_lba + iwlba*src_wbs_nlb, src_wbs_nlb, copy_buffer);
				p_dst->write_hdsector(p_dst->callback_data, p_dst->reading_offset + p_dst->part_lba +    bl*dst_wbs_nlb, dst_wbs_nlb, copy_buffer);
			}

			info->wlba_table[i] = wbfs_htons(bl);

			if(spinner)
				spinner(i,p_src->n_wbfs_sec_per_disc);
		}
	}
	else if (n_src_in_dst == 1)
	{
		int buf_dst_step = p_dst->wbfs_sec_sz;

		for( i=0; i< p_src->n_wbfs_sec_per_disc; i++)
		{
			unsigned int iwlba = wbfs_ntohs(d_src->header->wlba_table[i]);
			unsigned short bl = 0;
			if (iwlba)
			{
				p_src->read_hdsector( p_src->callback_data, p_src->reading_offset + p_src->part_lba + iwlba*src_wbs_nlb, src_wbs_nlb, copy_buffer);

				for( int l = 0; l < n_dst_in_src; ++l)
				{
					bl = alloc_block(p_dst);
					if (bl==0xffff)
						ERROR_WBFS("no space left on device (disc full)");

					p_dst->write_hdsector(p_dst->callback_data, p_dst->reading_offset + p_dst->part_lba + bl*dst_wbs_nlb, dst_wbs_nlb, copy_buffer + l * buf_dst_step);

					info->wlba_table[i*n_dst_in_src+l] = wbfs_htons(bl);
				}
			}

			if(spinner)
				spinner(i,p_src->n_wbfs_sec_per_disc);
		}
	}
	else if (n_dst_in_src == 1)
	{
		dst_copy_buffer = (unsigned char*)wbfs_ioalloc(p_dst->wbfs_sec_sz);

		if(!dst_copy_buffer)
			ERROR_WBFS("alloc memory");

		int buf_src_step = p_src->wbfs_sec_sz;
		bool need_to_write = false;

		for( i=0; i< p_src->n_wbfs_sec_per_disc; i++)
		{
			if(need_to_write && i%n_src_in_dst == 0)
			{
				unsigned short bl = 0;

				bl = alloc_block(p_dst);
				if (bl==0xffff)
					ERROR_WBFS("no space left on device (disc full)");

				p_dst->write_hdsector(p_dst->callback_data, p_dst->reading_offset + p_dst->part_lba +    bl*dst_wbs_nlb, dst_wbs_nlb, dst_copy_buffer);
				info->wlba_table[(i/n_src_in_dst)-1] = wbfs_htons(bl);

				memset(dst_copy_buffer, 0, p_dst->wbfs_sec_sz);
				need_to_write = false;
			}

			unsigned int iwlba = wbfs_ntohs(d_src->header->wlba_table[i]);
			if (iwlba)
			{
				p_src->read_hdsector( p_src->callback_data, p_src->reading_offset + p_src->part_lba + iwlba*src_wbs_nlb, src_wbs_nlb, dst_copy_buffer + (i%n_src_in_dst) * buf_src_step);
				need_to_write = true;
			}

			if(spinner)
				spinner(i,p_src->n_wbfs_sec_per_disc);
		}

		if(need_to_write)
		{
			unsigned short bl = 0;

			bl = alloc_block(p_dst);
			if (bl==0xffff)
				ERROR_WBFS("no space left on device (disc full)");

			p_dst->write_hdsector(p_dst->callback_data, p_dst->reading_offset + p_dst->part_lba +    bl*dst_wbs_nlb, dst_wbs_nlb, dst_copy_buffer);
			info->wlba_table[(i/n_src_in_dst)-1] = wbfs_htons(bl);

			memset(dst_copy_buffer, 0, p_dst->wbfs_sec_sz);
			need_to_write = false;
		}
	}

	if(spinner)
		spinner(p_src->n_wbfs_sec_per_disc,p_src->n_wbfs_sec_per_disc);

	// write disc info
	int disc_info_sz_lba = p_dst->disc_info_sz>>p_dst->hd_sec_sz_s;
	p_dst->write_hdsector(p_dst->callback_data,p_dst->reading_offset + p_dst->part_lba+1+discn*disc_info_sz_lba,disc_info_sz_lba,info);
	wbfs_sync(p_dst);
error:
	if(info)
		wbfs_iofree(info);
	if(copy_buffer)
		wbfs_iofree(copy_buffer);
	if(dst_copy_buffer)
		wbfs_iofree(dst_copy_buffer);


	// init with all free blocks

	return 0;
}

unsigned int wbfs_extract_file(wbfs_disc_t*d, char *path);
unsigned int wbfs_estimate_disc(wbfs_t *p, read_wiidisc_callback_t read_src_wii_disc, void *callback_data, partition_selector_t sel)
{
	unsigned char *b;
	int i;
	unsigned int tot;
	unsigned int wii_sec_per_wbfs_sect = 1 << (p->wbfs_sec_sz_s-p->wii_sec_sz_s);
	wiidisc_t *d = 0;
	unsigned char *used = 0;
	wbfs_disc_info_t *info = 0;

	tot = 0;

	used = (unsigned char*)wbfs_malloc(p->n_wii_sec_per_disc);
	if (!used)
	{
		ERROR_WBFS("unable to alloc memory");
	}

	d = wd_open_disc(read_src_wii_disc, callback_data);
	if (!d)
	{
		wbfs_message("Unable to open wii disc");
	}

	wd_build_disc_usage(d,sel,used);
	wd_close_disc(d);
	d = 0;

	info = (wbfs_disc_info_t*)wbfs_ioalloc(p->disc_info_sz);
	b = (unsigned char *)info;
	read_src_wii_disc(callback_data, 0, 0x100, info->disc_header_copy);

	fprintf(stderr, "estimating %c%c%c%c%c%c %s...\n",b[0], b[1], b[2], b[3], b[4], b[5], b + 0x20);

	for (i = 0; i < p->n_wbfs_sec_per_disc; i++)
	{
		if (block_used(used, i, wii_sec_per_wbfs_sect))
		{
			tot++;
		}
	}

error:
	if (d)
		wd_close_disc(d);

	if (used)
		wbfs_free(used);

	if (info)
		wbfs_iofree(info);

	return tot * ((p->wbfs_sec_sz / p->hd_sec_sz) * 512);
}

unsigned int wbfs_ren_disc(wbfs_t*p, unsigned char* discid, char* newname)
{
	wbfs_disc_t *d = wbfs_open_disc(p,discid);
	int disc_info_sz_lba = p->disc_info_sz>>p->hd_sec_sz_s;

	if(!d)
		return 1;

	memset(d->header->disc_header_copy+0x20, 0, 0x40);
	strncpy(d->header->disc_header_copy+0x20, newname, 0x39);

	p->write_hdsector(p->callback_data,p->reading_offset + p->part_lba+1+d->i*disc_info_sz_lba,disc_info_sz_lba,d->header);
	wbfs_close_disc(d);
	return 0;
}

// data extraction
unsigned int wbfs_extract_disc(wbfs_disc_t*d, rw_sector_callback_t write_dst_wii_sector,void *callback_data,progress_callback_t spinner)
{
	wbfs_t *p = d->p;
	unsigned char* copy_buffer = 0;
	unsigned int i;
	unsigned int src_wbs_nlb=p->wbfs_sec_sz/p->hd_sec_sz;
	unsigned int dst_wbs_nlb=p->wbfs_sec_sz/p->wii_sec_sz;

	copy_buffer = (unsigned char*)wbfs_ioalloc(p->wbfs_sec_sz);
	if(!copy_buffer)
		ERROR_WBFS("alloc memory");

	for( i=0; i< p->n_wbfs_sec_per_disc; i++)
	{
		unsigned int iwlba = wbfs_ntohs(d->header->wlba_table[i]);
		if (iwlba)
		{
			p->read_hdsector(p->callback_data, p->reading_offset + p->part_lba + iwlba*src_wbs_nlb, src_wbs_nlb, copy_buffer);
			
			write_dst_wii_sector(callback_data, i*dst_wbs_nlb, dst_wbs_nlb, copy_buffer);
		}

		if(spinner)
			spinner(i,p->n_wbfs_sec_per_disc);
	}

	if(spinner)
		spinner(p->n_wbfs_sec_per_disc,p->n_wbfs_sec_per_disc);

	wbfs_iofree(copy_buffer);
	return 0;
error:
	return 1;
}

// Extract disk CISO
unsigned int wbfs_extract_disc_ciso(wbfs_disc_t*d, rw_sector_callback_t write_dst_wii_sector,void *callback_data,progress_callback_t spinner)
{
	wbfs_t *p = d->p;
	unsigned char* copy_buffer = 0;
	unsigned int i;
	unsigned int src_wbs_nlb=p->wbfs_sec_sz/p->hd_sec_sz;
	unsigned int dst_wbs_nlb=p->wbfs_sec_sz/p->wii_sec_sz;
	unsigned int offset=0;

	copy_buffer = (unsigned char*)wbfs_ioalloc(p->wbfs_sec_sz);
	if(!copy_buffer)
		ERROR_WBFS("alloc memory");

	memset(copy_buffer,0,p->wbfs_sec_sz);
	copy_buffer[0]='C';
	copy_buffer[1]='I';
	copy_buffer[2]='S';
	copy_buffer[3]='O';

	i=dst_wbs_nlb<<15;

	copy_buffer[4]=i & 0xFF;
	copy_buffer[5]=(i>>8) & 0xFF;
	copy_buffer[6]=(i>>16) & 0xFF;
	copy_buffer[7]=(i>>24) & 0xFF;

	for( i=0; i< p->n_wbfs_sec_per_disc; i++)
	{
		if (wbfs_ntohs(d->header->wlba_table[i])) copy_buffer[i+8]=1;
		else copy_buffer[i+8]=0;
	}

	write_dst_wii_sector(callback_data, 0, 1, copy_buffer);

	offset=1;
	for( i=0; i< p->n_wbfs_sec_per_disc; i++)
	{
		unsigned int iwlba = wbfs_ntohs(d->header->wlba_table[i]);
		if (iwlba)
		{   
			if(spinner)
				spinner(i,p->n_wbfs_sec_per_disc);
			p->read_hdsector(p->callback_data, p->part_lba + iwlba*src_wbs_nlb, src_wbs_nlb, copy_buffer);
			write_dst_wii_sector(callback_data, offset, dst_wbs_nlb, copy_buffer);

			offset+=dst_wbs_nlb;
		}
	}

	wbfs_iofree(copy_buffer);
	return 0;
error:
	return 1;
}
